#!/usr/bin/env python3
# -*- coding: utf-8 -*-
##############################################################################
# IRKit command line interface
# -----------------------------------------------
# $Revision: 93 $
# $Id: irkit.py 93 2017-02-04 02:00:23Z  $
#
# Copyright (C) 2015-2017 Yoshihiro Naito. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
##############################################################################
import argparse
import json
import os
import sys
import urllib.request
from resolve_irkit import init_resolve, resolve_irkits
from socket import gethostbyname, gethostbyaddr

Version = '0.3.8'
__author__ = 'Y.Naito'
__date__ = '$Date: 2017-02-04 11:00:23 +0900 (2017/02/04 (土)) $'
__version__ = '$Id: irkit.py 93 2017-02-04 02:00:23Z  $'


def ir_save_dir():
    """Save directory of IR signal files"""
    return os.path.expanduser('~') + '/.irkit.d'


save_dir = ir_save_dir()  # Getした赤外線信号の保存先ディレクトリ名
make_filename = lambda signal_name: save_dir + '/signals/' + signal_name + '.json'


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Control IRKit')
    parser.add_argument('-v', '--version', action='version',
                        version='%(prog)s '+Version)
    parser.add_argument('-V', '--verbose', action='store_true',
                        default=False, help='verbose mode')
    subparsers = parser.add_subparsers(dest='command')
    # create the parser for the "get" command
    parser_get = subparsers.add_parser('get', aliases=['read'],
                                       help='GET command')
    parser_get.add_argument('ir_signal_name', action='store',
                            help='IR signal name')
    parser_get.add_argument('-a', '--address',
                            nargs=1,
                            action='store',
                            default=None,
                            dest='host',
                            metavar='address[:port]',
                            help='''IP address and port number of IRKit device.
Port 80 is default.''')
    # create the parser for the "post" command
    parser_post = subparsers.add_parser('post', aliases=['send'],
                                        help='POST command')
    parser_post.add_argument('ir_signal_name', action='store',
                             help='IR signal name')
    parser_post.add_argument('-a', '--address',
                             nargs=1,
                             action='store',
                             default=None,
                             dest='host',
                             metavar='address[:port]',
                             help='''IP address and port number of IRKit device.
Port 80 is default.''')
    # create the parser for the "show" command
    parser_show = subparsers.add_parser('show', aliases=['view'],
                                        help='Show IR signal data')
    parser_show.add_argument('ir_signal_name', action='store',
                             help='IR signal name')
    # create the parser for the "list" command
    parser_list = subparsers.add_parser(
                      'list',
                      help='Print list saved IR signal'
                  )
    # create the parser for the "rename" command
    parser_rename = subparsers.add_parser(
                        'rename',
                        aliases=['replace'],
                        help='Rename saved IR signal'
                    )
    parser_rename.add_argument('old_signal_name', action='store',
                               help='Old IR signal name')
    parser_rename.add_argument('new_signal_name', action='store',
                               help='New IR signal name')
    # create the parser for the "delete" command
    parser_delete = subparsers.add_parser(
                        'delete',
                        aliases=['remove'],
                        help='Delete saved IR signal'
                    )
    parser_delete.add_argument('ir_signal_name',
                               action='store',
                               help='IR signal name')
    # create the parser for the "update_devices" command
    parser_device = subparsers.add_parser(
                        'update_devices', aliases=['update'],
                        help='Update IRKit devices list'
                    )
    # create the parser for the "show_devices" command
    parser_device = subparsers.add_parser(
                        'show_devices',
                        help='Show IRKit devices list'
                    )
    #
    args = parser.parse_args()
    verbose = args.verbose
else:
    verbose = False


def print_message(msg):
    """Output error message to stderr"""
    if __name__ == '__main__':
        print(msg, file=sys.stderr)


class IRKit(object):
    """IRkit class"""
    def __init__(self, host, port):
        url = 'http://' + host + ':' + port + '/messages'
        self.request = urllib.request.Request(
                           url,
                           headers={'X-Requested-With': 'XMLHttpRequest'}
                       )

    def get(self):
        """最も新しい受信した赤外線信号を返す。"""
        with urllib.request.urlopen(url=self.request) as self.f:
            return self.f.read().decode('utf-8')

    def post(self, json_data):
        """赤外線信号を送信する。"""
        with urllib.request.urlopen(url=self.request,
                                    data=json_data.encode('utf-8')):
            pass


class NoIRKitError(Exception):
    """No IRKit device Exception"""
    pass


class IRKitNotFoundError(Exception):
    """IRKit device is not found Exception"""
    pass


def check_dir(path):
    """pathディレクトリの有無を調べ、無い場合は作成する。"""
    if os.path.exists(path):
        assert os.path.isdir(path), path + ' is not directory.'
    else:
        os.mkdir(path)


def search_irkit():
    """BonjourでIRKitを探し、見つかったIRKitの情報リストを返却する。"""
    init_resolve()
    irkits = resolve_irkits()
    assert irkits, 'No IRKit device'
    devices = []
    for irkit in irkits:
        host, address, port = (irkit[0],
                               gethostbyname(irkit[0]),
                               str(irkit[1]))
        devices.append({'Host': host,
                        'IPaddress': address,
                        'Port': port})
#   print('search IRKit devices: {0}'.format(devices))
    return devices


def save_devices(devices):
    """Save IRKit devices info"""
#   print('save_devices: {0}'.format(devices))
    save_file = save_dir + '/irkit.save'
    with open(save_file, mode='wt', encoding='utf-8') as f:
        for dev in devices:
            json.dump(dev, f)
            if dev != devices[-1]:
                f.write('\n')


def load_devices():
    """Load IRKit devices info"""
    save_file = save_dir + '/irkit.save'
    devices = []
    with open(save_file, mode='rt', encoding='utf-8') as f:
        for line in f:
            devices.append(json.loads(line))
    if not devices:
        raise NoIRKitError
#   print('loaded IRKit devices: {0}'.format(devices))
    return devices


def get_irkit_info():
    """
    保存しておいた、または検出した IRKit の情報辞書(IPアドレス,
    ポート番号,サーバ名)を順番に返す。
    """
    save_file = save_dir + '/irkit.save'
    try:
        devices = load_devices()
    except (FileNotFoundError,
            json.decoder.JSONDecodeError, NoIRKitError):
        devices = search_irkit()
        save_devices(devices)
    else:
        # Check IRKit info.
        ip_changed = False
        for (i, irkit_info) in enumerate(devices):
            try:
                ip_addr = gethostbyname(irkit_info['Host'])
            except:
                continue  # IRKitが見えなくなっても情報は保持する
            if irkit_info['IPaddress'] != ip_addr:
                # IP address has changed
                devices[i]['IPaddress'] = ip_addr
                ip_changes = True
        if ip_changed:
            save_devices(devices)
    for irkit_info in devices:
        if verbose:
            print('IRKit device: {0}'.format(irkit_info))
        yield irkit_info


def update_devices():
    """IRKitデバイスを検索し、情報をファイルに保存する。"""
    save_devices(search_irkit())


def show_devices():
    """IRKitデバイスを検索し、情報を表示する。"""
    for irkit_info in search_irkit():
        print('"{0}" {1}:{2}'.format(irkit_info['Host'],
                                     irkit_info['IPaddress'],
                                     irkit_info['Port']))


def get_irkit_hostname(address):
    """指定したIPアドレスのIRKitホスト名を返却する。"""
    for info in get_irkit_info():
        if address == info['IPaddress']:
            return info['Host']
    return None


def get(address, port, host, signal_name):
    """GET sub-command"""
#   print('get(address={0}, port={1}, host={2}, signal_name={3})'.format(address, port, host, signal_name))

    irkit = IRKit(address, port)
    data = irkit.get()
    if data:
        ir_data = json.loads(data)
        ir_data['name'] = signal_name
        ir_data['hostname'] = host.split('.')[0]
        ir_data['type'] = 'single'
        file_name = make_filename(signal_name)
        if verbose:
            print('{0}:{1} <= {2}:{3}'.format(file_name, ir_data, host, address))
        with open(file_name, mode='wt', encoding='utf-8') as f:
            json.dump(ir_data, f, indent=2, ensure_ascii=False,
                      separators=(',', ': '))
        return True
    else:
        return False


def post(address, port, signal_name):
    """POST sub-command"""
#   print('post(address={0}, port={1}, signal_name={2})'.format(address, port, signal_name))

    file_name = make_filename(signal_name)
    with open(file_name, mode='rt', encoding='utf-8') as f:
        ir_data = json.load(f)
        if address is None:
            # 赤外線ファイル内に保存されているホスト名のIPアドレスを
            # 求める。
            hostname = ir_data['hostname']
            for info in get_irkit_info():
                info_host = info['Host'].split('.')[0]
                if (info_host == hostname.lower()):
                    address = info['IPaddress']
                    break
            else:
                print_message('Host name "{0}" of IRKit is not found.'.format(hostname))
                raise IRKitNotFoundError
        if verbose:
            print('{0}:{1} => {2}'.format(file_name, ir_data, address))
        irkit = IRKit(address, port)
        irkit.post(json.dumps(ir_data))


def show(signal_name):
    """Show sub-command"""
    file_name = make_filename(signal_name)
    with open(file_name, mode='rt', encoding='utf-8') as f:
        print('{0} = {1}'.format(file_name, json.load(f)))


def list(root):
    """List sub-command"""
    for path, dirs, files in os.walk(root):
        for file in sorted(files):
            if file[0] != '.':            # 先頭が'.'のファイルは除く
                if verbose:
                    print(os.path.join(path, file))
                else:
                    print(os.path.join(
                            path,
                            os.path.splitext(file)[0]).replace(root, '')
                         )


def rename(old_name, new_name):
    """Rename sub-command"""
    if old_name != new_name:
        old_file_name = make_filename(old_name)
        assert os.path.exists(old_file_name), \
            'The IR signal ' + old_name + ' does not exist.'
        if verbose:
            print('rename '+old_name+' to '+new_name)
        with open(old_file_name, mode='rt', encoding='utf-8') as fin:
            ir_data = json.load(fin)
            ir_data['name'] = new_name
            with open(make_filename(new_name),
                      mode='wt', encoding='utf-8') as fout:
                json.dump(ir_data,
                          fout,
                          indent=2,
                          ensure_ascii=False,
                          separators=(',', ': '))
                os.remove(old_file_name)
    else:
        pass


def delete(signal_name):
    """Delete sub-command"""
    file_name = make_filename(signal_name)
    assert os.path.exists(file_name), \
        'The IR signal ' + signal_name + ' does not exist.'
    if verbose:
        print('delete '+file_name)
    os.remove(file_name)


def main():
    check_dir(save_dir)
    check_dir(save_dir + '/signals')

    if args.command == 'get' or args.command == 'read':
        if args.host:
            address = args.host[0].split(':')[0]
            try:
                port = args.host[0].split(':')[1]
            except IndexError:
                port = '80'
            hostname = get_irkit_hostname(address)
            if hostname is None:
                print_message('IRKit with IP address "{0}" is not found.'.format(address))
                return
            if not get(address, port, hostname, args.ir_signal_name):
                print_message('No data')
        else:
            for info in get_irkit_info():
                if get(info['IPaddress'], info['Port'],
                       info['Host'], args.ir_signal_name):
                    break
            else:
                print_message('No data')
    elif args.command == 'post' or args.command == 'send':
        if args.host:
            address = args.host[0].split(':')[0]
            try:
                port = args.host[0].split(':')[1]
            except IndexError:
                port = '80'
            hostname = get_irkit_hostname(address)
            if hostname is None:
                print_message('IRKit with IP address "{0}" is not found.'.format(address))
                return
            post(address, port, args.ir_signal_name)
        else:
            post(None, '80', args.ir_signal_name)
    elif args.command == 'show' or args.command == 'view':
        show(args.ir_signal_name)
    elif args.command == 'list':
        list(save_dir + '/signals')
    elif args.command == 'rename' or args.command == 'replace':
        rename(args.old_signal_name, args.new_signal_name)
    elif args.command == 'delete' or args.command == 'remove':
        delete(args.ir_signal_name)
    elif args.command == 'update_devices' or args.command == 'update':
        update_devices()
    elif args.command == 'show_devices':
        show_devices()
    else:
        pass


if __name__ == '__main__':
    main()
